﻿-- Author      : Nogarder
-- Create Date : 2013/5/4 15:04:07
--register initialize function
--NPA.DataCollector.RegisterModule( datatype, initfunc, uninitfunc )
local CoolDownTimer
NPA.DataCollector.RegisterModule("CoolDown",
	function()
		--global
		local UnitGUID=UnitGUID
		local table_insert=table.insert
		local table_getn=table.getn
		local floor=floor
		local NPA_Timer_AddTimer=NPA.Timer.AddTimer
		local NPA_Timer_RemoveTimer=NPA.Timer.RemoveTimer
		--main --using local for speed
		local NPA_EventHandler_FireEvent=NPA.EventHandler.FireEvent
		local NPA_SpellInfo_GetSpellInfo=NPA.SpellInfo.GetSpellInfo
		local NPA_DataCollector_GetUnitDataByGUID=NPA.DataCollector.GetUnitDataByGUID
		--
		--create or show timer
		if(not CoolDownTimer)then
			CoolDownTimer=NPA.Timer.CreateTimer(function(unitdata)
				NPA_EventHandler_FireEvent(7,"CoolDown_Updated",unitdata)
			end)
		else
			CoolDownTimer:Show()
		end
		--macro for insert cooldown to data base on unitbuff/unitdebuff
		local NPA_CoolDown_SortCoolDown=function(cooldownlist, spellid, casterguid, appliedtime,endtime,spelltype)
			local max=table_getn(cooldownlist)
			--check data
			for i=1,max do
				local cooldown=cooldownlist[i]
				--check with applied time, it's accurate
				if(cooldown.SpellId==spellid and cooldown.AppliedTime==appliedtime)then
					--already exists ,check refresh
					--check expiration time and spelltype
					--ignore type check if spelltype==nil
					if(cooldown.EndTime==endtime)then
						return i
					else
						cooldown.CasterGUID=casterguid
						cooldown.EndTime=endtime or cooldown.EndTime
						cooldown.SpellType=spelltype
						return i
					end
					--in case of refreshing, there's only one spellid per caster on same unit
				elseif(cooldown.SpellId==spellid and cooldown.CasterGUID==casterguid)then
					cooldown.CasterGUID=cooldown.CasterGUID or casterguid
					cooldown.EndTime=endtime or cooldown.EndTime
					cooldown.SpellType=spelltype
					return i
				end
			end
			--not exists, try use a unused data slot
			for i=1,max do
				local cooldown=cooldownlist[i]
				if(cooldown.SpellId==nil)then
					cooldown.SpellId=spellid
					cooldown.CasterGUID=casterguid
					cooldown.EndTime=endtime
					cooldown.AppliedTime=appliedtime
					cooldown.SpellType=spelltype
					return i
				end
			end
			--all used ,insert a new one
			local i=max+1
			cooldownlist[i]={
				SpellId=spellid,
				CasterGUID=casterguid,
				EndTime=endtime,
				AppliedTime=appliedtime,
				SpellType=spelltype
			}
			return i
		end
		--
		local NPA_CoolDown_UpdateCoolDown=function(sourceunitdata,cooldownList,affactlist)
			local max=table_getn(cooldownlist)
			for spellid,affact in pairs(affactlist)do
				for i=1,max do
					local cooldown=cooldownlist[i]
					if(cooldown.SpellId==spellid)then
						--0=reduce cooldown 1=extend cooldown ,2=reduce duration,3=extend duration
						if(affact.AffactType==0)then
							if(affact.AffactAmout==0)then
								--remove timer
								NPA_Timer_RemoveTimer(CoolDownTimer,sourceunitdata,i)
								--remove data
								cooldown.SpellId=nil
							else
								--reduce duration
								cooldown.EndTime=cooldown.EndTime-affact.AffactAmount
								--refresh timer
								NPA_Timer_AddTimer(CoolDownTimer,sourceunitdata,i,cooldown.EndTime)
							end
						elseif(affact.AffactType==1 and affact.AffactAmount~=0)then
							cooldown.EndTime=cooldown.EndTime+affact.AffactAmount
							--refresh timer
							NPA_Timer_AddTimer(CoolDownTimer,sourceunitdata,i,cooldown.EndTime)
						end
					end
				end
			end
		end
		--combatlog event filter
		local NPA_CoolDown_CombatLogEventFilter={
			["SPELL_AURA_APPLIED"]=1,
			["SPELL_AURA_REFRESH"]=1, --same as apply
			["SPELL_DAMAGE"]=1,--same as apply
			["SPELL_MISSED"]=1,
			["SPELL_CAST_SUCCESS"]=1,
			--["SPELL_CAST_START"]=1, --start doesnt count
			--
			["SPELL_AURA_REMOVED"]=2, 
			["SPELL_AURA_BROKEN"]=2, --count as remove
			["SPELL_AURA_BROKEN_SPELL"]=2, --count as remove
			["SPELL_STOLEN"]=2, --count as remove
			["SPELL_DISPEL"]=2, --count as remove
		}
		--
		local bit_band=bit.band
		local COMBATLOG_OBJECT_REACTION_MASK=COMBATLOG_OBJECT_REACTION_MASK
		--combatlog event functions
		local NPA_CoolDown_CombatLogEventFunc={
			-- [1] for cooldown on used
			--destunitdata,arg[1],arg[4],arg[12]
			[1]=function(sourceunitdata,appliedtime,casterguid,spellid,spelltype)
				--create a data set if there isn't one
				if(not sourceunitdata.CoolDown)then
					sourceunitdata.CoolDown={}
				end
				--
				local spelldata=NPA_SpellInfo_GetSpellInfo(spellid)
				if(spelldata.CoolDown)then
					local cooldowns=sourceunitdata.CoolDown
					if(spelldata.CoolDown>0 and spelldata.CoolDownType==0)then -- type=0 is cd on used
						local endtime=appliedtime+spelldata.CoolDown
						--sort it
						local index=NPA_CoolDown_SortCoolDown(cooldowns, spellid, casterguid, appliedtime,endtime,spelltype)
						NPA_Timer_AddTimer(CoolDownTimer,sourceunitdata,index,endtime)
					end
					--also check affacted spell
					if(spelldata.AffactSpellList)then
						NPA_CoolDown_UpdateCoolDown(sourceunitdata,cooldowns,spelldata.AffectSpellList)
					end
				end
			end,
			--[2] for cooldown on removed
			[2]=function(sourceunitdata,appliedtime,casterguid,spellid,spelltype)
				--create a data set if there isn't one
				if(not sourceunitdata.CoolDown)then
					sourceunitdata.CoolDown={}
				end
				--
				local spelldata=NPA_SpellInfo_GetSpellInfo(spellid)
				if(spelldata.CoolDown)then
					local cooldowns=sourceunitdata.CoolDown
					if(spelldata.CoolDown>0 and spelldata.CoolDownType==1)then -- type=1 is cd on removed
						local endtime=appliedtime+spelldata.CoolDown
						--sort it
						NPA_CoolDown_SortCoolDown(cooldowns, spellid, casterguid, appliedtime,endtime,spelltype)
						NPA_Timer_AddTimer(CoolDownTimer,sourceunitdata,index,endtime)
					end
					--also check affacted spell
					if(spelldata.AffactSpellList)then
						NPA_CoolDown_UpdateCoolDown(sourceunitdata,cooldowns,spelldata.AffectSpellList)
					end
				end
			end,
		}
		--get 'any' event then filter it locally
		NPA.DataCollector.RegisterCombatLogEvent("Any","DataCollector_CoolDown",function(arg,sourceunitdata,destunitdata)
			--[1]=timestamp
			--[2]=event
			--[3]=hidecaster
			--[4]=sourceGUID
			--[5]=srouceName
			--[6]=sourceFlags
			--[7]=sourceFlags2 raidflag
			--[8]=destGUID
			--[9]=destName
			--[10]=destFlags
			--[11]=destFlags2 raidflag
			--[12]=spellid
			--[13]=spellname
			--[14]=spellschool
			--[15]=spelltype (Buff/Debuff)
			--[16]=amount
			--check with the filter			--return if destGUID not exists or spellid is nil
			if(NPA_CoolDown_CombatLogEventFilter[arg[2]] and arg[4]~=0 and arg[12])then
				--run the pointed function
				NPA_CoolDown_CombatLogEventFunc[NPA_CoolDown_CombatLogEventFilter[arg[2]]](sourceunitdata,arg[1],arg[4],arg[12],arg[15])
				--fire event
				NPA_EventHandler_FireEvent(7,"CoolDown_Updated",sourceunitdata)
			end
		end)
	end,
	function()
		--hide timer
		if(CoolDownTimer)then
			CoolDownTimer:Hide()
		end
		--unregister events
		NPA.DataCollector.RegisterCombatLogEvent("Any","DataCollector_CoolDown",nil)
	end
	--no data transfer as frame-based won't get any cooldown
	)
