﻿-- Author      : Nogarder
-- Create Date : 2012/11/22 14:48:27
local pairs=pairs
local table_insert=table.insert
local table_getn=table.getn
local table_sort=table.sort
local mod=mod
local NPA_GetTime=NPA.GetTime
local NPA_RunFunc=NPA.RunFunc
local floor=floor
--
local NPA_SpellInfo_GetSpellInfo=NPA.SpellInfo.GetSpellInfo

local NPA_FontManager_ManageFont2=NPA.FontManager.ManageFont2

local NPA_NumberManager_CreateNumber=NPA.NumberManager.CreateNumber
local NPA_NumberManager_ManageNumber=NPA.NumberManager.ManageNumber
local NPA_NumberManager_SetNumber=NPA.NumberManager.SetNumber
local NPA_NumberManager_SetNumberColor=NPA.NumberManager.SetNumberColor

local NPA_BarManager_CreateBarGroup=NPA.BarManager.CreateBarGroup
local NPA_BarManager_ManageBarGroup=NPA.BarManager.ManageBarGroup
local NPA_BarManager_SetBarGroup=NPA.BarManager.SetBarGroup
local NPA_BarManager_SetBarGroupColor=NPA.BarManager.SetBarGroupColor

local NPA_TextureManager_ManageTexture2=NPA.TextureManager.ManageTexture2

local NPA_Color_GetColorBySpellType=NPA.Color.GetColorBySpellType
--
NPA.IconGroupManager={}
--
local NPA_IconGroupManager_DirectionList={
	[1]={ --grow right then up
		Anchor="BOTTOMLEFT",
		TAnchor="BOTTOMRIGHT",
		DirX=1,
		DirY=1,
	},
	[2]={ -- grow left then up
		Anchor="BOTTOMRIGHT",
		TAnchor="BOTTOMLEFT",
		DirX=-1,
		DirY=1,
	},
	[3]={ -- grow right then down
		Anchor="TOPLEFT",
		TAnchor="TOPRIGHT",
		DirX=1,
		DirY=-1,
	},
	[4]={ --grow left then down
		Anchor="TOPRIGHT",
		TAnchor="TOPLEFT",
		DirX=-1,
		DirY=-1,
	},
	[5]={ -- grow up then right
		Anchor="BOTTOMLEFT",
		TAnchor="TOPLEFT",
		DirX=1,
		DirY=1,
	},
	[6]={ --grow up then left
		Anchor="BOTTOMRIGHT",
		TAnchor="TOPRIGHT",
		DirX=-1,
		DirY=1,
	},
	[7]={ -- grow down then right
		Anchor="TOPLEFT",
		TAnchor="BOTTOMLEFT",
		DirX=1,
		DirY=-1,
	},
	[8]={ -- grow down then left
		Anchor="TOPRIGHT",
		TAnchor="BOTTOMRIGHT",
		DirX=-1,
		DirY=-1,
	},
}
--
local NPA_IconGroupManager_CreateIconFrame=function(mainframe)
	local iconframe=CreateFrame("Frame",nil,mainframe)
	iconframe:SetPoint("CENTER",mainframe,"CENTER",0,0)
	iconframe:EnableMouse(false)
	--
	return iconframe
end
--
local NPA_IconGroupManager_BarGroupUpdateFunc=function(module,length,data)
	NPA_BarManager_SetBarGroup(module,length/data.Length)
end
--
local NPA_IconGroupManager_NumberUpdateFunc=function(module,length,data)
	NPA_NumberManager_SetNumber(module,length,data.Length,true)
end
--
local NPA_IconGroupManager_CreateModule=function(moduletype,icongroup)
	local moduleframe
	if(moduletype=="Icon" or moduletype=="Texture")then
		--create icon
		moduleframe=CreateFrame("Frame",nil,icongroup)
		moduleframe:EnableMouse(false)
		local texture=moduleframe:CreateTexture()
		texture:SetPoint("TOPLEFT",moduleframe,"TOPLEFT",0,0)
		texture:SetPoint("BOTTOMRIGHT",moduleframe,"BOTTOMRIGHT",0,0)
		moduleframe.Texture=texture
	elseif(moduletype=="Text")then
		--create text
		moduleframe=CreateFrame("Frame",nil,icongroup)
		moduleframe:EnableMouse(false)
		local text=moduleframe:CreateFontString()
		text:SetPoint("CENTER",moduleframe,"CENTER",0,0)
		moduleframe.Text=text
	elseif(moduletype=="Number")then
		moduleframe=NPA_NumberManager_CreateNumber(icongroup)
	elseif(moduletype=="BarGroup")then
		moduleframe=NPA_BarManager_CreateBarGroup(icongroup)
	end
	--
	moduleframe.ModuleType=moduletype
	return moduleframe
end
--
local NPA_IconGroupManager_AddModule=function(icon,modulesbase,setting,data,icongroup)
	--get a unused one fom icongroup
	local max=table_getn(modulesbase)
	local module
	local moduletype=setting.ModuleType
	--
	for i=1,max do
		local modulesbasei=modulesbase[i]
		if(modulesbasei.ModuleType==moduletype and not modulesbasei:IsShown())then
			module=modulesbasei
			break
		end
	end
	if(not module)then
		--create a new one
		modulesbase[max+1]=NPA_IconGroupManager_CreateModule(moduletype,icongroup)
		module=modulesbase[max+1]
	end
	--
	module:SetParent(icon)
	--
	module:ClearAllPoints()
	local anchordata=setting.AnchorData
	module:SetPoint(anchordata.From,icon,anchordata.To,anchordata.OffsetX,anchordata.OffsetY)
	--
	if(moduletype=="Icon")then
		module:SetWidth(setting.Width)
		module:SetHeight(setting.Height)
		module.Texture:SetTexture(NPA_SpellInfo_GetSpellInfo(data.Spellid).Icon)
		module.UpdateFunc=nil
	elseif(moduletype=="Texture")then
		module:SetWidth(setting.Width)
		module:SetHeight(setting.Height)
		NPA_TextureManager_ManageTexture2(module.Texture,setting.ProfileName)
		if(setting.ColorMode==1)then
			local color=NPA_Color_GetColorBySpellType(data.SpellType)
			module.Texture:SetVertexColor(color.R,color.G,color.B,color.A)
		end
		module.UpdateFunc=nil
	elseif(moduletype=="Text")then
		NPA_FontManager_ManageFont2(module.Text,setting)
		module.Text:SetText(NPA_SpellInfo_GetSpellInfo(data.Spellid).Name)
		module:SetWidth(module.Text:GetWidth())
		module:SetHeight(module.Text:GetHeight())
		if(setting.ColorMode==1)then
			local color=NPA_Color_GetColorBySpellType(data.SpellType)
			module.Text:SetTextColor(color.R,color.G,color.B,color.A)
		end
		module.UpdateFunc=nil
	elseif(moduletype=="Number")then
		NPA_NumberManager_ManageNumber(module,setting)
		NPA_NumberManager_SetNumber(module,0,0)
		--check data type
		if(setting.DataType=="Time")then
			module.UpdateFunc=NPA_IconGroupManager_NumberUpdateFunc
		else
			if(setting.DataType=="Stack" and data.Count and data.Count>1)then
				--set stack
				local count=data.Count or 1
				local maxstack=NPA_SpellInfo_GetSpellInfo(data.Spellid).MaxStack
				if(not maxstack or maxstack<count)then
					maxstack=count
				end
				NPA_NumberManager_SetNumber(module,count,maxstack)
			else
				module:Hide() --as hiden, return directly
				return module
			end
			module.UpdateFunc=nil
		end
		if(setting.ColorMode==1)then
			local color=NPA_Color_GetColorBySpellType(data.SpellType)
			NPA_NumberManager_SetNumberColor(module,color)
		end
	elseif(moduletype=="BarGroup")then
		NPA_BarManager_ManageBarGroup(module,setting)
		--check data type
		if(setting.DataType=="Time")then
			module.UpdateFunc=NPA_IconGroupManager_BarGroupUpdateFunc
		else
			if(setting.DataType=="Stack" and data.Count and data.Count>1)then
				--set stack
				local count=data.Count
				local maxstack=NPA_SpellInfo_GetSpellInfo(data.Spellid).MaxStack
				if(not maxstack or maxstack<count)then
					maxstack=count
				end
				NPA_BarManager_SetBarGroup(module,count/maxstack)
			else
				module:Hide() --as hiden, return directly
				return module
			end
			module.UpdateFunc=nil
		end
		if(setting.ColorMode==1)then
			local color=NPA_Color_GetColorBySpellType(data.SpellType)
			NPA_BarManager_SetBarGroupColor(module,color)
		end
	end
	--set module
	module:SetScale(setting.Scale)
	module:SetAlpha(setting.Alpha)
	module:SetFrameLevel(setting.FrameLevel)
	--
	module:Show()
	--
	return module
end
--
local NPA_IconGroupManager_SetIcon=function(icon,icongroup,settings,data)
	icon:SetWidth(settings.Width)
	icon:SetHeight(settings.Height)
	icon:SetScale(settings.Scale)
	icon:SetAlpha(settings.Alpha)
	icon:SetFrameLevel(settings.FrameLevel)
	--
	local modules
	if(not icon.Modules)then
		--create new
		icon.Modules={}
		modules=icon.Modules
	else
		--clear old
		modules=icon.Modules
		for index,_ in pairs(modules)do
			modules[index]=nil
		end
	end
	--set module
	local modulesbase=icongroup.Modules
	local i=1
	for _,setting in pairs(settings.Modules)do
		modules[i]=NPA_IconGroupManager_AddModule(icon,modulesbase,setting,data,icongroup)
		i=i+1
	end
end
--update group display
NPA.IconGroupManager.UpdateGroup=function(icongroup)
	--sort priority
	local nTime=NPA_GetTime()
	local iconlist={}
	local i=1
	for index,data in pairs(icongroup.DataList)do
		--ignore old one and those out of timelimit
		--don't check end time as it always have one
		if(data.Spellid and (icongroup.TimeLimit==0 or (data.EndTime-nTime)<icongroup.TimeLimit))then
			iconlist[i]=data
			i=i+1
		end
	end
	--
	table_sort(iconlist,function(a,b)
		--sort by priority and end time
		if(a.Priority~=b.Priority)then
			--priority from large to small
			return a.Priority>b.Priority
		else
			--tme from small to large
			return a.EndTime<b.EndTime
		end
	end)
	--set icons
	local max=icongroup.Max
	local count=table_getn(iconlist)
	--
	if(count<1)then
		--nothing to show
		icongroup:Hide()
		return
	end
	--
	if(max>count)then
		max=count
	end
	local icons=icongroup.Icons
	--hide modules
	for _,module in pairs(icongroup.Modules)do
		module:Hide()
	end
	--
	local offsety=0
	local offsetx=0
	local rowheight=0
	local rowwidth=0
	local perrow=icongroup.PerRow
	--anchor and direction
	local directioninfo=NPA_IconGroupManager_DirectionList[icongroup.Direction]
	local anchor=directioninfo.Anchor
	local tanchor=directioninfo.TAnchor
	local dirx=directioninfo.DirX
	local diry=directioninfo.DirY
	local gapx=icongroup.GapX
	local gapy=icongroup.GapY
	--
	--grow left/right then up/down
	if(icongroup.Direction<5)then
		for i=1,max do
			if(not icons[i])then
				icons[i]=NPA_IconGroupManager_CreateIconFrame(icongroup)
			end
			local icon=icons[i]
			local data=iconlist[i]
			local setting=data.Setting
			--set data
			data.Showing=icon
			icon.Data=data
			--
			local height=setting.Height*setting.Scale+gapy
			local width=setting.Width*setting.Scale+gapx
			--clear points
			icon:ClearAllPoints()
			--
			if(mod(i,perrow)==1)then --the first one of every row
				--caculate row width and height
				if(rowwidth>offsetx)then
					offsetx=rowwidth
				end
				offsety=offsety+rowheight
				rowwidth=width
				rowheight=height
				--set point
				icon:SetPoint(anchor,icongroup,anchor,gapx*dirx,(offsety+gapy)*diry)
			else
				--caculate row width and height
				if(height>rowheight)then
					rowheight=height
				end
				rowwidth=rowwidth+width
				--set point
				icon:SetPoint(anchor,icons[i-1],tanchor,gapx*dirx,0)
			end
			--set icon
			NPA_IconGroupManager_SetIcon(icon,icongroup,setting,data)
			icon:Show()
		end
		--add the current row
		offsety=offsety+rowheight
	else
		--grow up/down then left/right
		for i=1,max do
			if(not icons[i])then
				icons[i]=NPA_IconGroupManager_CreateIconFrame(icongroup)
			end
			local icon=icons[i]
			local data=iconlist[i]
			local setting=data.Setting
			--
			data.Showing=icon
			icon.Data=data
			--
			local height=setting.Height*setting.Scale+gapy
			local width=setting.Width*setting.Scale+gapx
			--clear points
			icon:ClearAllPoints()
			--
			if(mod(i,perrow)==1)then --the first one of every row
				--caculate row width and height
				if(rowheight>offsety)then
					offsety=rowheight
				end
				offsetx=offsetx+rowwidth
				rowwidth=width
				rowheight=height
				--set point
				icon:SetPoint(anchor,icongroup,anchor,(offsetx+gapx)*dirx,gapy*diry)
			else
				--caculate row width and height
				if(width>rowwidth)then
					rowwidth=width
				end
				rowheight=rowheight+height
				--set point
				icon:SetPoint(anchor,icons[i-1],tanchor,0,gapy)
				icon:Show()
			end
			--set icon
			NPA_IconGroupManager_SetIcon(icon,icongroup,setting,data)
			icon:Show()
			--
		end
		--add the current row
		offsetx=offsetx+rowwidth
	end
	--hide unused icons
	for i=max+1,table_getn(icons)do
		icons[i]:Hide()
	end
	--
	if(offsetx<rowwidth)then
		offsetx=rowwidth
	end
	if(offsety<rowheight)then
		offsety=rowheight
	end
	--
	icongroup:SetWidth(offsetx)
	icongroup:SetHeight(offsety)
	icongroup:Show()
end
--add a new icon
NPA.IconGroupManager.InsertIcon=function(icongroup,starttime,endtime,spellid,spelltype,count,iconsetting)
	--
	local datalist=icongroup.DataList
	--
	local max=table_getn(datalist)
	for i=1,max do
		local data=datalist[i]
		--reuse old one to save resources
		--check spellid
		if(data.Spellid==nil)then
			data.Spellid=spellid
			data.StartTime=starttime
			data.EndTime=endtime
			data.Setting=iconsetting
			data.Showing=nil
			data.SpellType=spelltype
			data.Count=Count
			--also caculate length 
			data.Length=endtime-starttime
			return
		elseif(data.StartTime==starttime and data.EndTime==endtime and data.Spellid==spellid)then
			--if these match then it's the same one, do nothing then return
			return
		end
	end
	--all used up ,add a new one
	datalist[max+1]={
		Spellid=spellid,
		StartTime=starttime,
		EndTime=endtime,
		Setting=iconsetting,
		Showing=nil,
		SpellType=spelltype,
		Count=Count,
		Length=endtime-starttime
	}
end
--
local NPA_IconGroupManager_UpdateGroup=NPA.IconGroupManager.UpdateGroup
--
NPA.IconGroupManager.ManageIconGroup=function(icongroup,groupsetting)
	icongroup.Direction=groupsetting.Direction
	icongroup.PerRow=groupsetting.PerRow
	icongroup.Max=groupsetting.Max
	icongroup.TimeLimit=groupsetting.TimeLimit
	icongroup.GapX=groupsetting.GapX
	icongroup.GapY=groupsetting.GapY
	NPA_IconGroupManager_UpdateGroup(icongroup)
end
--
NPA.IconGroupManager.ClearIconGroup=function(icongroup)
	--reuse old ones
	local datalist=icongroup.DataList
	for i=1,table_getn(datalist)do
		datalist[i].Spellid=nil --remove spellid to set it unused
	end
	local icons=icongroup.Icons
	for i=1,table_getn(icongroup.Icons)do
		icons[i]:Hide()
	end
	--clear modules
	local modules=icongroup.Modules
	for i=1,table_getn(modules)do
		modules[i]:Hide()
	end
end
--on hide,set width and height to 1
local NPA_IconGroupManager_GroupOnHide=function(icongroup)
	icongroup:SetWidth(1)
	icongroup:SetHeight(1)
end
--onupdate check icon timer
local NPA_IconGroupManager_GroupOnUpdate=function(icongroup)
	local nTime=NPA_GetTime()
	local update=false
	local datalist=icongroup.DataList
	local limit=icongroup.TimeLimit
	for i=1,table_getn(datalist)do
		local data=datalist[i]
		if(data.Spellid)then --if no spellid then not used
			local length=data.EndTime-nTime
			if(length<0)then
				data.Spellid=nil
				update=true
			elseif(data.Showing==nil)then
				if(length<limit)then
					--check not showing but in time limit
					update=true
				end
			else
				--update timer text
				for _,module in pairs(data.Showing.Modules)do
					NPA_RunFunc(module.UpdateFunc,module,length,data)
				end
			end
		end
	end
	--only update once,this should save some resources
	if(update)then
		NPA_IconGroupManager_UpdateGroup(icongroup)
	end
end
--
NPA.IconGroupManager.CreateIconGroup=function(mainframe)
	local icongroup=CreateFrame("Frame",nil,mainframe)
	--icongroup:SetPoint("CENTER",mainframe,"CENTER",0,0)
	icongroup:EnableMouse(false)
	--
	icongroup.DataList={}
	icongroup.Icons={}
	icongroup.Modules={}
	--
	icongroup:SetScript("OnUpdate",NPA_IconGroupManager_GroupOnUpdate)
	icongroup:SetScript("OnHide",NPA_IconGroupManager_GroupOnHide)
	--
	return icongroup
end
